import os
import pathlib
import re
from xml.dom import minidom

import requests

from ...utils import color


def SAFE(
    productid: str,
    output: str,
    runchecks: bool = True,
    quiet: bool = False,
    method: str = "requests",
):
    """Download a Sentinel-2 image using a product ID.

    Args:
        productid (str): The full id of the Sentinel-2 product.
        output (str): The output directory to save the image.
        runchecks (bool, optional): If True, run checks on the gsutil. Only used if method is "gsutil". Defaults to True.
        quiet (bool, optional): If True, do not print info messages. Defaults to False.
        method (str, optional): The method to use to download the image. Defaults to "requests". Available options are "requests" and "gsutil".

    Returns:
        str: The path to the downloaded image.

    Raises:
        Exception: If the productid is not valid.
        Exception: If gsutil is not installed.

    Examples:
        >>> import maskay
        >>> s2idfull = "S2A_MSIL1C_20190212T142031_N0207_R010_T19FDF_20190212T191443"
        >>> s2idpath = maskay.download.s2.SAFE(s2idfull, "/home/user/")
    """

    # Get tile grid info.
    rgx_expr = "(.*)_(.*)_(.*)_(.*)_(.*)"
    tile_info = re.search(rgx_expr, productid, re.IGNORECASE).group(4)[1:]
    p01 = tile_info[0:2]
    p02 = tile_info[2]
    p03 = tile_info[3:5]

    # Download the S2 L1C image folder.
    if method == "gsutil":
        # check if gsutil is installed
        if runchecks:
            __gsutilcheck()

        # 3. Create the path to download the image.
        base_uri = "gs://gcp-public-data-sentinel-2/tiles"
        BASEPATH = "%s/%s/%s/%s/%s.SAFE/" % (base_uri, p01, p02, p03, productid)
        if quiet:
            os.system("gsutil -q cp -r %s %s" % (BASEPATH, output))
        else:
            print("Running: gsutil -m cp -r %s %s" % (BASEPATH, output))
            os.system("gsutil cp -r %s %s" % (BASEPATH, output))

    if method == "requests":
        base_uri = "https://storage.googleapis.com/gcp-public-data-sentinel-2/tiles"
        folder_img = "%s/%s/%s/%s/%s.SAFE/" % (base_uri, p01, p02, p03, productid)

        # Create the output folder if it does not exist.
        PATH = pathlib.Path(output)
        if not PATH.exists():
            PATH.mkdir(parents=True, exist_ok=True)

        # download MTD_MSIL1C.xml
        mtd_msil1c_xml = "%sMTD_MSIL1C.xml" % folder_img
        r = requests.get(mtd_msil1c_xml, allow_redirects=True)

        # write a xml file
        BASEPATH = pathlib.Path((PATH / productid).as_posix() + ".SAFE")
        BASEPATH.mkdir(exist_ok=True)
        S2xmlFile = BASEPATH / "MTD_MSIL1C.xml"
        with open(S2xmlFile, "wb") as f:
            f.write(r.content)

        # read a XML file
        file = minidom.parse(S2xmlFile.as_posix())

        # download images
        models = file.getElementsByTagName("IMAGE_FILE")

        # if models is empty return an exception
        if len(models) == 0:
            raise Exception(
                color.RED
                + "No images found. Please check the"
                + color.BOLD
                + " productid"
                + color.END
                + "."
            )

        # Download all S2L1C bands
        for model in models:

            # Get S2 image band from google cloud storage
            S2todownload = (folder_img + model.firstChild.data) + ".jp2"
            band = re.search(
                string=pathlib.Path(model.firstChild.data).stem,
                pattern="(.*)_(.*)_(.*)",
                flags=re.IGNORECASE,
            ).group(3)

            # NO download the TCI image if it is exists.
            if band == "TCI":
                continue

            if not quiet:
                print(
                    color.BLUE
                    + color.BOLD
                    + "Downloading [%s]: " % band
                    + color.END
                    + color.UNDERLINE
                    + S2todownload
                    + color.END
                )

            # where to save the S2 image
            S2tosave = BASEPATH / model.firstChild.data

            # Create the folder if it does not exist.
            S2tosave.parent.mkdir(parents=True, exist_ok=True)
            S2tosave = S2tosave.as_posix() + ".jp2"

            # Download S2 images
            r = requests.get(S2todownload, allow_redirects=True)
            with open(S2tosave, "wb") as f:
                f.write(r.content)
    return BASEPATH.as_posix()


def __gsutilcheck():
    """Check if the gsutil is installed.

    Raises:
        Exception: If the gsutil is not installed.
        Exception: If the gsutil is not working.

    Returns:
        bool: True if gsutil is installed.
    """

    # check if gsutil is installed silent
    if not os.system("gsutil --version > /dev/null 2>&1") == 0:
        raise Exception("gsutil is not installed")

    # check if gsutil is working
    # if not os.system("gsutil ls > /dev/null 2>&1") == 0:
    #    raise Exception("gsutil is not working")
    return True



from .safe import SAFE


def SAFEGEE(
    s2id: str,
    output: str,
    runchecks: bool = True,
    quiet: bool = False,
    method: str = "requests",
):
    """Download a Sentinel-2 image using Google Earth Engine IDs.

    Args:
        s2id (str): The GEE Sentinel-2 ID.
        output (str): The output directory.
        runchecks (bool, optional): If True, run checks on the Earth Engine and gsutil. Defaults to True.
        quiet (bool, optional): If True, do not print info messages. Defaults to False.
        method (str, optional): The method to use to download the image. Defaults to "requests". Available options are "requests" and "gsutil".

    Raises:
        Exception: If the Earth Engine is not working.

    Returns:
        str: The path to the downloaded image.

    Examples:
        >>> import ee
        >>> import maskay
        >>> ee.Initialize()
        >>> s2id = "20190212T142031_20190212T143214_T19FDF"
        >>> s2idpath = maskay.download.s2.SAFEGEE(s2id, "/home/user/")
    """

    try:
        import ee
    except ImportError:
        print("Please install the following packages: earthengine-api.")

    # check if ee is initialized
    if runchecks:
        __eecheck()

    # Identify S2 image using GEE S2 ID
    s2 = ee.ImageCollection("COPERNICUS/S2_HARMONIZED")
    s2 = s2.filter(ee.Filter.eq("system:index", s2id)).first()

    # Get tile information.
    productid = s2.get("PRODUCT_ID").getInfo()

    # Download S2 image.
    BASEPATH = SAFE(productid, output, runchecks, quiet, method)

    return BASEPATH


def __eecheck():
    """Simple check to see if the Earth Engine is working.
    Raises:
        Exception: If the Earth Engine is not working.
    Returns:
        bool: True if gsutil is installed.
    """
    try:
        ee.Image(0)
    except:
        raise Exception("ee is not initialized")

    return True



import numpy as np
import gc

try:
    import torch
except ImportError:
    raise ImportError("Please install the following packages: torch.")

from maskay.module import MaskayModule

class Module(MaskayModule):
    def __init__(
        self,
        cropsize: int = 512,
        overlap: int = 32,
        device: str = "cpu",
        batchsize: int = 1,
        order: str = "BCHW",
        quiet: int = False,
    ):
        super().__init__(cropsize, overlap, device, batchsize, order, quiet)

    def inProcessing(self, tensor: np.ndarray):
        pass

    def forward(self, x):
        pass

    def outProcessing(self, tensor: np.ndarray):
        pass

    def _run(self, tensor: np.ndarray):
        tensor = torch.Tensor(tensor)

        # Run the model
        with torch.no_grad():
            if self.device == "cuda":
                tensor = tensor.cuda()
            tensor = self.forward(tensor).detach().cpu().numpy()
            torch.cuda.empty_cache()
            gc.collect()
        return tensor


import pathlib

import gdown
import numpy as np

from maskay.torch import Module
from maskay.utils import get_models_path, softmax

class UnetMobV2(Module):
    def __init__(self):
        super().__init__()
        self.model = model_setup()

    def forward(self, x):
        return self.model(x)
    
    def inProcessing(self, tensor: np.ndarray):
        # If all the pixels are zero skip the run and outProcessing.
        if np.sum(tensor) == -999:
            shp = tensor.shape
            tensor = np.zeros(
                (shp[0], 4, shp[2], shp[3])
            ) # 4 is the number of the output classes
            # replace first channel by ones
            tensor[:, 0, :, :] = 10000
            return [tensor]
        return tensor / 10000

    def outProcessing(self, tensor: np.ndarray):
        return (softmax(tensor, axis=1) * 10000).astype(np.int16)


def model_setup():
    # Check if packages are installed
    is_external_package_installed = []

    try:
        import pytorch_lightning as pl
    except ImportError:
        is_external_package_installed.append("pytorch_lightning")

    try:
        import segmentation_models_pytorch as smp
    except ImportError:
        is_external_package_installed.append("segmentation_models_pytorch")

    if is_external_package_installed != []:
        nopkgs = ', '.join(is_external_package_installed)
        raise ImportError(
            f"Please install the following packages: {nopkgs}."
        )    
    
    
    class UnetMobV2Class(pl.LightningModule):
        def __init__(self):
            super().__init__()
            self.model = smp.Unet(
                encoder_name="mobilenet_v2",
                encoder_weights=None,
                in_channels=13,
                classes=4,
            )

        def forward(self, x):
            return self.model(x)
        
    filename = pathlib.Path(get_models_path()) / "unetmobv2.ckpt"
    # Download the model if it doesn't exist
    if not filename.is_file():
        # download file using gdown
        url = "https://drive.google.com/uc?id=1o9LeVsXCeD2jmS-G8s7ZISfciaP9v-DU"
        gdown.download(url, filename.as_posix())
    # Load the model
    model = UnetMobV2Class().load_from_checkpoint(filename.as_posix())
    model.eval()
    return model



import numpy as np

try:
    import tensorflow as tf
except ImportError:
    raise ImportError("Please install the following packages: tensorflow.")

from maskay.module import MaskayModule


class Module(MaskayModule):
    def __init__(
        self,
        cropsize: int = 512,
        overlap: int = 32,
        device: str = "cpu",
        batchsize: int = 1,
        order: str = "BCWH",
        quiet: int = False,
    ):
        super().__init__(cropsize, overlap, device, batchsize, order, quiet)

    def inProcessing(self, tensor: np.ndarray):
        pass

    def forward(self, x):
        pass

    def outProcessing(self, tensor: np.ndarray):
        pass

    def _run(self, tensor: np.ndarray):
        if self.device == "cuda":
            with tf.device("/GPU:0"):
                tensor = np.moveaxis(self.forward(tensor).numpy(), 3, 1)
        elif self.device == "cpu":
            with tf.device("/CPU:0"):
                tensor = np.moveaxis(self.forward(tensor).numpy(), 3, 1)
        else:
            raise ValueError("Unknown device: " + str(self.device))
        return tensor



import pathlib
import re
from collections import OrderedDict
from typing import Dict, List, Union

import numpy as np
import requests


# Beautiful print --------------------------------------------------------------
class color:
    """Color class for printing colored text."""

    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


# Setup auxiliary functions ---------------------------------------------------
def get_models_path():
    """ Get the path to the cloud detection models."""

    # Path to save the models
    cred_path = pathlib.Path(
        "~/.config/maskaymodels/",
    ).expanduser()

    # create the folder if it does not exist
    if not cred_path.is_dir():
        cred_path.mkdir(parents=True, exist_ok=True)

    return cred_path.as_posix()


def download_file_from_google_drive(id, destination):
    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params={"id": id}, stream=True)
    token = get_confirm_token(response)

    if token:
        params = {"id": id, "confirm": token}
        response = session.get(URL, params=params, stream=True)

    save_response_content(response, destination)


def get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith("download_warning"):
            return value

    return None


def save_response_content(response, destination):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)


# Processing auxiliary functions ----------------------------------------------
def softmax(X, theta=1.0, axis=None):
    """
    Compute the softmax of each element along an axis of X.

    Parameters
    ----------
    X: ND-Array. Probably should be floats.
    theta (optional): float parameter, used as a multiplier
        prior to exponentiation. Default = 1.0
    axis (optional): axis to compute values along. Default is the
        first non-singleton axis.

    Returns an array the same size as X. The result will sum to 1
    along the specified axis.
    """

    # make X at least 2d
    y = np.atleast_2d(X)

    # find axis
    if axis is None:
        axis = next(j[0] for j in enumerate(y.shape) if j[1] > 1)

    # multiply y against the theta parameter,
    y = y * float(theta)

    # subtract the max for numerical stability
    y = y - np.expand_dims(np.max(y, axis=axis), axis)

    # exponentiate y
    y = np.exp(y)

    # take the sum along the specified axis
    ax_sum = np.expand_dims(np.sum(y, axis=axis), axis)

    # finally: divide elementwise
    p = y / ax_sum

    # flatten if X was 1D
    if len(X.shape) == 1:
        p = p.flatten()

    return p


# Maskay dict object ----------------------------------------------------------
def ListFiles(path, pattern=None, full_names=False, recursive=False):
    """list.file (like in R) function"""
    files = list(list_file_gen(path, pattern, full_names, recursive))
    files_str = [str(file) for file in files]
    files_str.sort()
    return files_str


def list_file_gen(path, pattern=None, full_names=False, recursive=False):
    """List files like in R - generator"""
    path = pathlib.Path(path)
    for file in path.iterdir():
        if file.is_file():
            if pattern is None:
                if full_names:
                    yield file
                else:
                    yield file.name
            elif pattern is not None:
                regex_cond = re.compile(pattern=pattern)
                if regex_cond.search(str(file)):
                    if full_names:
                        yield file
                    else:
                        yield file.name
        elif recursive:
            yield from ListFiles(file, pattern, full_names, recursive)


def OrderBands(files: List[str], sensor: str = "Sentinel-2"):
    if sensor == "Sentinel-2":
        bandS2 = {
            "Aerosol": "_B01.jp2",
            "Blue": "_B02.jp2",
            "Green": "_B03.jp2",
            "Red": "_B04.jp2",
            "RedEdge1": "_B05.jp2",
            "RedEdge2": "_B06.jp2",
            "RedEdge3": "_B07.jp2",
            "NIR": "_B08.jp2",
            "NIR2": "_B8A.jp2",
            "WaterVapor": "_B09.jp2",
            "Cirrus": "_B10.jp2",
            "SWIR1": "_B11.jp2",
            "SWIR2": "_B12.jp2",
        }

        S2r = {}
        for key, value in bandS2.items():
            for file in files:
                if re.search(value, file):
                    S2r[key] = file
                    continue
        return MaskayObject(S2r)
    else:
        raise ValueError("The sensor is not supported yet!")


def dictreverser(dict, ref):
    refnames = np.array(list(ref.keys()))
    dictnames = np.array(list(dict.keys()))
    condition = ~np.isin(dictnames, refnames)
    if np.any(condition):
        raise ValueError(
            "The band names '%s' are invalid" % ", ".join(dictnames[condition])
        )
    dictreverser = {}
    for dictname in dictnames:
        dictreverser[str(ref[dictname])] = dictname
    return dictreverser


class MaskayObject:
    def __init__(self, dict: Dict[str, str]) -> None:
        self.ref = OrderedDict(
            {
                "Aerosol": 0,
                "Blue": 1,
                "Green": 2,
                "Red": 3,
                "RedEdge1": 4,
                "RedEdge2": 5,
                "RedEdge3": 6,
                "NIR": 7,
                "NIR2": 8,
                "WaterVapor": 9,
                "Cirrus": 10,
                "SWIR1": 11,
                "SWIR2": 12,
                "TIR1": 13,
                "TIR2": 14,
                "HV": 15,
                "VH": 16,
                "HH": 17,
                "VV": 18,
            }
        )
        self.ref_rev = OrderedDict({v: k for k, v in self.ref.items()})
        self.dict = OrderedDict(dict)

    def to_dict(self):
        return self.dict

    def to_list(self):
        return list(self.dict.values())

    def __getitem__(self, index):
        # If is a range object, convert it to a list
        if isinstance(index, range):
            index = list(index)
        elif isinstance(index, int):
            band = self.ref_rev[index]
            return MaskayObject({band: self.dict[band]})
        elif isinstance(index, str):
            band = self.dict[index]
            return MaskayObject({index: band})
        elif isinstance(index, slice):
            # Obtain the bands between the start and stop slice
            container_list = list()
            for key in self.ref_rev.keys():
                if key >= index.start and key <= index.stop:
                    container_list.append(key)

            # Using the indexes of the bands, create a new dictionary
            container_dict = {}
            for element in container_list:
                band = self.ref_rev[element]
                container_dict[band] = self.dict[band]
            return MaskayObject(container_dict)

        if isinstance(index, list):
            container_dict = OrderedDict()
            for element in index:
                if isinstance(element, int):
                    band = self.ref_rev[element]
                    container_dict[band] = self.dict[band]
                elif isinstance(element, str):
                    container_dict[element] = self.dict[element]
            return MaskayObject(container_dict)
        else:
            raise ValueError("Subset index is not valid")

    def __repr__(self):
        MAXSTRING = 30  # Maximum string length to show
        ESPACE = 15  # Establish a constant space between the elements

        msg = ""
        for key, value in self.dict.items():
            # Add three dots if the string is longer than MAXSTRING
            if len(value) > MAXSTRING:
                value_string = value[0:30] + " ... " + value[-30:-1] + value[-1]
            else:
                value_string = value

            # Constant space between the elements
            white_space = " " * (ESPACE - len("%s [%02d]" % (key, 0)))
            name = color.BLUE + color.BOLD + key + color.END
            order = color.BLUE + color.BOLD + "%02d" % self.ref[key] + color.END

            # Print it!
            msg += "%s %s[%s]: %s\n" % (name, white_space, order, value_string)
        return msg


def MaskayDict(
    path: Union[str, pathlib.Path],
    pattern: str = None,
    full_names: bool = False,
    recursive: bool = False,
    sensor: str = "Sentinel-2",
) -> MaskayObject:
    """Create a MaskayDict object from a path.

    Args:
        path (Union[str, pathlib.Path]): Path to the folder containing the files.
        pattern (str, optional): Pattern to match the files. Defaults to None.
        full_names (bool, optional): If True, return the full path of the files. Defaults to False.
        recursive (bool, optional): If True, recursively list the files. Defaults to False.
        sensor (str, optional): Sensor name. Defaults to "Sentinel-2".

    Returns:
        MaskayObject: MaskayDict
    """
    # List all the files according to a pattern
    files: List[str] = ListFiles(
        path=path, pattern=pattern, full_names=full_names, recursive=recursive
    )

    return OrderBands(files, sensor=sensor)



import numpy as np
import rasterio.features

from pyproj import Transformer
from pyproj.aoi import AreaOfInterest
from pyproj.database import query_utm_crs_info

S2BANDS_L2A = [
    "B01",
    "B02",
    "B03",
    "B04",
    "B05",
    "B06",
    "B07",
    "B08",
    "B8A",
    "B09",
    "B11",
    "B12",
    "SCL",
    "AOT",
    "WVP",
]


def PC(coords, iniDate, endDate, buffer, bands=None):
    # Check if packages are installed
    is_external_package_installed = []

    try:
        import planetary_computer as pc
    except ImportError:
        is_external_package_installed.append("planetary_computer")

    try:
        import stackstac
    except ImportError:
        is_external_package_installed.append("stackstac")

    try:
        import pystac_client
    except ImportError:
        is_external_package_installed.append("pystac_client")

    if is_external_package_installed != []:
        nopkgs = ', '.join(is_external_package_installed)
        raise ImportError(
            f"Please install the following packages: {nopkgs}."
        )

    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(coords[0], coords[1], coords[0], coords[1]),
    )
    epsg = int(utm_crs_list[0].code)

    aoi = towerFootprint(coords[0], coords[1], buffer)
    bbox = rasterio.features.bounds(towerFootprint(coords[0], coords[1], buffer, False))

    CATALOG = pystac_client.Client.open(
        "https://planetarycomputer.microsoft.com/api/stac/v1"
    )

    SEARCH = CATALOG.search(
        intersects=aoi, datetime=f"{iniDate}/{endDate}", collections=["sentinel-2-l2a"]
    )

    items = list(SEARCH.get_items())

    if bands is None:
        bands = S2BANDS_L2A

    S2 = signAndStack(items, bbox, epsg, bands)

    return S2


# Auxiliary functions ------------------------------------------------------
def signAndStack(items, bbox, epsg, bands):
    signed_items = []

    for item in items:
        item.clear_links()
        signed_items.append(pc.sign(item).to_dict())

    S2 = stackstac.stack(
        signed_items,
        assets=bands,
        resolution=10,
        bounds=bbox,
        epsg=epsg,
    ).where(
        lambda x: x > 0, other=np.nan
    )  # NO DATA IS ZERO -> THEN TRANSFORM ZEROS TO NO DATA

    return S2


def towerFootprint(x, y, distance, latlng=True, resolution=10):
    utm_crs_list = query_utm_crs_info(
        datum_name="WGS 84",
        area_of_interest=AreaOfInterest(x, y, x, y),
    )
    transformer = Transformer.from_crs(
        "EPSG:4326", "EPSG:" + utm_crs_list[0].code, always_xy=True
    )
    inverse_transformer = Transformer.from_crs(
        "EPSG:" + utm_crs_list[0].code, "EPSG:4326", always_xy=True
    )
    newCoords = transformer.transform(x, y)

    newCoords = [round(i / resolution) * resolution for i in newCoords]

    E = newCoords[0] + distance
    W = newCoords[0] - distance
    N = newCoords[1] + distance
    S = newCoords[1] - distance

    polygon = [
        [W, S],
        [E, S],
        [E, N],
        [W, N],
        [W, S],
    ]

    if latlng:
        polygon = [list(inverse_transformer.transform(x[0], x[1])) for x in polygon]

    footprint = {
        "type": "Polygon",
        "coordinates": [polygon],
    }

    return footprint



import pathlib
import gdown

import numpy as np

from maskay.tensorflow import Module
from maskay.utils import get_models_path


class KappaModelUNetL2A(Module):
    def __init__(self):
        super().__init__()
        self.model = model_setup()

    def forward(self, x):
        return self.model(x)

    def inProcessing(self, tensor: np.ndarray):
        # If all the pixels are zero skip the run and outProcessing.
        if np.sum(tensor) == 0:
            shp = tensor.shape
            tensor = np.zeros(
                (shp[0], 6, shp[1], shp[2])
            ) # 6 is the number of the output classes
            return [tensor]
        
        factor = np.array(
            [0.003906309604644775, 0.2833447754383087, 0.27905699610710144]
            + [0.2668497860431671, 0.2558480203151703, 0.250980406999588]
            + [0.24597543478012085, 0.24182498455047607, 0.24206912517547607]
            + [0.2535133957862854, 0.2232242375612259, 0.23524834215641022]
            + [0.07557793706655502]
        )
        return tensor / 65_635 / factor[:, None, None]

    def outProcessing(self, tensor: np.ndarray):
        return (tensor * 10000).astype(np.int16)


def model_setup():
    
    CLASSES_KAPPAZETA = [
        "UNDEFINED",
        "CLEAR",
        "CLOUD_SHADOW",
        "SEMI_TRANSPARENT_CLOUD",
        "CLOUD",
        "MISSING",
    ]
        
    # Check if packages are installed
    is_external_package_installed = []

    try:
        import tensorflow as tf
    except ImportError:
        is_external_package_installed.append("tensorflow")

    if is_external_package_installed != []:
        nopkgs = ', '.join(is_external_package_installed)
        raise ImportError(
            f"Please install the following packages: {nopkgs}."
        )

    class KappaModelUNet:
        def __init__(self, version: str = "L1C"):
            if version == "L1C":
                self.features = list(range(0, 13))            
            if version == "L2A":
                self.features = list(range(0, 14))            
            self.classes = CLASSES_KAPPAZETA        

        def construct(self, width=None, height=None):
            """
            Construct the model.
            """
            # For symmetrical neighbourhood, width and height must be odd numbers.
            input_shape = (width, height, len(self.features))

            with tf.name_scope("Model"):
                inputs = tf.keras.layers.Input(input_shape, name="input")

                conv1 = tf.keras.layers.Conv2D(
                    64, 3, activation="relu", padding="same", kernel_initializer="he_normal"
                )(inputs)
                conv1 = tf.keras.layers.Conv2D(
                    64, 3, activation="relu", padding="same", kernel_initializer="he_normal"
                )(conv1)
                pool1 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv1)

                conv2 = tf.keras.layers.Conv2D(
                    128,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(pool1)
                conv2 = tf.keras.layers.Conv2D(
                    128,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(conv2)
                pool2 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv2)

                conv3 = tf.keras.layers.Conv2D(
                    256,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(pool2)
                conv3 = tf.keras.layers.Conv2D(
                    256,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(conv3)
                pool3 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(conv3)

                conv4 = tf.keras.layers.Conv2D(
                    512,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(pool3)
                conv4 = tf.keras.layers.Conv2D(
                    512,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(conv4)
                drop4 = tf.keras.layers.Dropout(0.5)(conv4)
                pool4 = tf.keras.layers.MaxPool2D(pool_size=(2, 2))(drop4)

                conv5 = tf.keras.layers.Conv2D(
                    1024,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(pool4)
                conv5 = tf.keras.layers.Conv2D(
                    1024,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(conv5)
                drop5 = tf.keras.layers.Dropout(0.5)(conv5)

                up6 = tf.keras.layers.Conv2D(
                    512,
                    2,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(tf.keras.layers.UpSampling2D(size=(2, 2))(drop5))
                merge6 = tf.keras.layers.concatenate([drop4, up6], axis=3)
                conv6 = tf.keras.layers.Conv2D(
                    512,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(merge6)
                conv6 = tf.keras.layers.Conv2D(
                    512,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(conv6)

                up7 = tf.keras.layers.Conv2D(
                    256,
                    2,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(tf.keras.layers.UpSampling2D(size=(2, 2))(conv6))
                merge7 = tf.keras.layers.concatenate([conv3, up7], axis=3)
                conv7 = tf.keras.layers.Conv2D(
                    256,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(merge7)
                conv7 = tf.keras.layers.Conv2D(
                    256,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(conv7)

                up8 = tf.keras.layers.Conv2D(
                    128,
                    2,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(tf.keras.layers.UpSampling2D(size=(2, 2))(conv7))
                merge8 = tf.keras.layers.concatenate([conv2, up8], axis=3)
                conv8 = tf.keras.layers.Conv2D(
                    128,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(merge8)
                conv8 = tf.keras.layers.Conv2D(
                    128,
                    3,
                    activation="relu",
                    padding="same",
                    kernel_initializer="he_normal",
                )(conv8)

                up9 = tf.keras.layers.Conv2D(
                    64, 2, activation="relu", padding="same", kernel_initializer="he_normal"
                )(tf.keras.layers.UpSampling2D(size=(2, 2))(conv8))
                merge9 = tf.keras.layers.concatenate([conv1, up9], axis=3)

                conv9 = tf.keras.layers.Conv2D(
                    64, 3, activation="relu", padding="same", kernel_initializer="he_normal"
                )(merge9)
                conv9 = tf.keras.layers.Conv2D(
                    64, 3, activation="relu", padding="same", kernel_initializer="he_normal"
                )(conv9)
                conv9 = tf.keras.layers.Conv2D(
                    2, 3, activation="relu", padding="same", kernel_initializer="he_normal"
                )(conv9)
                conv10 = tf.keras.layers.Conv2D(
                    len(self.classes), (1, 1), activation="softmax"
                )(conv9)

                self.model = tf.keras.Model(inputs, conv10)

                return self.model
            
    filename = pathlib.Path(get_models_path()) / "kappamaskl1c.hdf5"
    
    # Download the model if it doesn't exist
    if not filename.is_file():
        # download file using gdown
        url = "https://drive.google.com/uc?id=1D2YxjbVETzmUcrv560exB4khFbiSfeqA"
        gdown.download(url, filename.as_posix())
    # Load the model
    model = KappaModelUNet("L2A").construct()
    model.load_weights(filename.as_posix())
    
    return model



from maskay.tensorsat import TensorSat


class Predictor:
    def __init__(
        self,
        cropsize: int = 512,
        overlap: int = 32,
        device: str = "cpu",
        batchsize: int = 1,
        order: int = "BCHW",
        quiet: int = False,
    ):
        self.batchsize = batchsize
        self.cropsize = cropsize
        self.overlap = overlap
        self.device = device
        self.order = order
        self.quiet = quiet
        self.result = None

    def predict(self, model, tensor: TensorSat):
        model.batchsize = self.batchsize
        model.cropsize = self.cropsize
        model.overlap = self.overlap
        model.device = self.device
        model.order = self.order
        model.quiet = self.quiet
        if model.device == "cuda":
            model.model = model.model.cuda()
        
        self.result = model._predict(tensor=tensor)
        return self.result

    def __repr__(self) -> str:
        # no long line
        msg = (
            "Predictor("
            + "cropsize="
            + str(self.cropsize)
            + ", "
            + "overlap="
            + str(self.overlap)
            + ", "
            + "device="
            + str(self.device)
            + ", "
            + "order="
            + str(self.order)
            + "batchsize="
            + str(self.batchsize)
            + ", "
            + "quiet="
            + str(self.quiet)
            + ")"
        )

        return msg

    def __str__(self) -> str:
        return self.__repr__()



import zipfile
import pathlib
import gdown

import numpy as np

from maskay.tensorflow import Module
from maskay.utils import get_models_path, softmax

class DynamicWorld(Module):
    def __init__(self):
        super().__init__()
        self.model = model_setup()

    def forward(self, x):
        return self.model(x)
    
    def inProcessing(self, tensor: np.ndarray) -> np.ndarray:
        # If all the pixels are zero skip the run and outProcessing.
        if np.sum(tensor) == 0:
            shp = tensor.shape
            tensor = np.zeros(
                (shp[0], 9, shp[1], shp[2])
            ) # 9 is the number of the output classes
            return [tensor]
        
        NORM_PERCENTILES = np.array(
            [
                [1.7417268007636313, 2.023298706048351],
                [1.7261204997060209, 2.038905204308012],
                [1.6798346251414997, 2.179592821212937],
                [1.7734969472909623, 2.2890068333026603],
                [2.289154079164943, 2.6171674549378166],
                [2.382939712192371, 2.773418590375327],
                [2.3828939530384052, 2.7578332604178284],
                [2.1952484264967844, 2.789092484314204],
                [1.554812948247501, 2.4140534947492487]            
            ]
        )
        tensor = np.log(tensor * 0.005 + 1)
        tensor = (tensor - NORM_PERCENTILES[:, 0]) / NORM_PERCENTILES[:, 1]
        tensor = np.exp(tensor * 5 - 1)
        tensor = tensor / (tensor + 1)
        return tensor.astype("float32")

    def outProcessing(self, tensor: np.ndarray) -> np.ndarray:
        return (softmax(tensor) * 10000).astype(np.int16)
    
# random int
def model_setup():
    # Check if packages are installed
    is_external_package_installed = []

    try:
        import tensorflow as tf
    except ImportError:
        is_external_package_installed.append("tensorflow")

    if is_external_package_installed != []:
        nopkgs = ', '.join(is_external_package_installed)
        raise ImportError(
            f"Please install the following packages: {nopkgs}."
        )

    filename = (pathlib.Path(get_models_path()) / "dynamicworld.zip")

    # Download the model if it doesn't exist
    if not filename.is_file():
        # download file using gdown
        url = "https://drive.google.com/uc?id=1DcvM2AvkJB0qBzymp-MBgNtYcO47ILBF"
        gdown.download(url, filename.as_posix(), quiet=False)
    
    # if a zip file
    if filename.suffix == ".zip" and filename.is_file():
        with zipfile.ZipFile(filename, "r") as zip_ref:
            zip_ref.extractall(filename.parent)
    FolderName = str(filename.parent / "DynamicWorld")
    
    return tf.saved_model.load(FolderName)



from collections import OrderedDict
from typing import Union

import numpy as np
import rioxarray
import xarray as xr
from affine import Affine
from rasterio.enums import Resampling

from .utils import color


class TensorSat:
    def __init__(
        self,
        cache=False,
        align=False,
        Aerosol=None,
        Blue=None,
        Green=None,
        Red=None,
        RedEdge1=None,
        RedEdge2=None,
        RedEdge3=None,
        NIR=None,
        NIR2=None,
        WaterVapor=None,
        Cirrus=None,
        SWIR1=None,
        SWIR2=None,
        TIR1=None,
        TIR2=None,
        HV=None,
        VH=None,
        HH=None,
        VV=None,
    ):
        self.cache: bool = cache
        self.align: bool = align
        self.Aerosol: Union[str, np.ndarray, xr.DataArray] = Aerosol
        self.Blue: Union[str, np.ndarray, xr.DataArray] = Blue
        self.Green: Union[str, np.ndarray, xr.DataArray] = Green
        self.Red: Union[str, np.ndarray, xr.DataArray] = Red
        self.RedEdge1: Union[str, np.ndarray, xr.DataArray] = RedEdge1
        self.RedEdge2: Union[str, np.ndarray, xr.DataArray] = RedEdge2
        self.RedEdge3: Union[str, np.ndarray, xr.DataArray] = RedEdge3
        self.NIR: Union[str, np.ndarray, xr.DataArray] = NIR
        self.NIR2: Union[str, np.ndarray, xr.DataArray] = NIR2
        self.WaterVapor: Union[str, np.ndarray, xr.DataArray] = WaterVapor
        self.Cirrus: Union[str, np.ndarray, xr.DataArray] = Cirrus
        self.SWIR1: Union[str, np.ndarray, xr.DataArray] = SWIR1
        self.SWIR2: Union[str, np.ndarray, xr.DataArray] = SWIR2
        self.TIR1: Union[str, np.ndarray, xr.DataArray] = TIR1
        self.TIR2: Union[str, np.ndarray, xr.DataArray] = TIR2
        self.HV: Union[str, np.ndarray, xr.DataArray] = HV
        self.VH: Union[str, np.ndarray, xr.DataArray] = VH
        self.HH: Union[str, np.ndarray, xr.DataArray] = HH
        self.VV: Union[str, np.ndarray, xr.DataArray] = VV
        self.ref = OrderedDict(
            {
                "Aerosol": 0,
                "Blue": 1,
                "Green": 2,
                "Red": 3,
                "RedEdge1": 4,
                "RedEdge2": 5,
                "RedEdge3": 6,
                "NIR": 7,
                "NIR2": 8,
                "WaterVapor": 9,
                "Cirrus": 10,
                "SWIR1": 11,
                "SWIR2": 12,
                "TIR1": 13,
                "TIR2": 14,
                "HV": 15,
                "VH": 16,
                "HH": 17,
                "VV": 18,
            }
        )
        self.ref_rev = OrderedDict({v: k for k, v in self.ref.items()})
        self.dictarray = None
        self.bounds = None

    def to_xarray(self, object: Union[str, np.ndarray] = None) -> xr.DataArray:
        """From string and numpy array to xr.DataArray.

        Args:
            object (Union[str, np.ndarray], optional): Object to convert.
                Defaults to None.

        Raises:
            ValueError: If the object is not a string or a numpy array.

        Returns:
            xr.DataArray: A xr.DataArray object.
        """
        if isinstance(object, str):
            with rioxarray.open_rasterio(object, "r") as src:
                raster = np.squeeze(src, axis=0)
                if self.cache:
                    raster = raster.compute()
                return raster

        elif isinstance(object, np.ndarray):
            coords = [list(range(d)) for d in object.shape]
            dims = ["y", "x"]
            return (
                xr.DataArray(object, coords=coords, dims=dims)
                .rio.write_nodata(-999)
                .rio.write_transform(Affine(1.0, 0.0, -0.5, 0.0, 1.0, -0.5))
            )
        elif isinstance(object, xr.DataArray):
            return object
        elif object is None:
            return None
        else:
            raise ValueError("object must be a string, or an array (numpy, xarray).")

    @property
    def Aerosol(self):
        return self._Aerosol

    @Aerosol.setter
    def Aerosol(self, value):
        self._Aerosol = self.to_xarray(value)

    @property
    def Blue(self):
        return self._Blue

    @Blue.setter
    def Blue(self, value):
        self._Blue = self.to_xarray(value)

    @property
    def Green(self):
        return self._Green

    @Green.setter
    def Green(self, value):
        self._Green = self.to_xarray(value)

    @property
    def Red(self):
        return self._Red

    @Red.setter
    def Red(self, value):
        self._Red = self.to_xarray(value)

    @property
    def RedEdge1(self):
        return self._RedEdge1

    @RedEdge1.setter
    def RedEdge1(self, value):
        self._RedEdge1 = self.to_xarray(value)

    @property
    def RedEdge2(self):
        return self._RedEdge2

    @RedEdge2.setter
    def RedEdge2(self, value):
        self._RedEdge2 = self.to_xarray(value)

    @property
    def RedEdge3(self):
        return self._RedEdge3

    @RedEdge3.setter
    def RedEdge3(self, value):
        self._RedEdge3 = self.to_xarray(value)

    @property
    def NIR(self):
        return self._NIR

    @NIR.setter
    def NIR(self, value):
        self._NIR = self.to_xarray(value)

    @property
    def NIR2(self):
        return self._NIR2

    @NIR2.setter
    def NIR2(self, value):
        self._NIR2 = self.to_xarray(value)

    @property
    def WaterVapor(self):
        return self._WaterVapor

    @WaterVapor.setter
    def WaterVapor(self, value):
        self._WaterVapor = self.to_xarray(value)

    @property
    def Cirrus(self):
        return self._Cirrus

    @Cirrus.setter
    def Cirrus(self, value):
        self._Cirrus = self.to_xarray(value)

    @property
    def SWIR1(self):
        return self._SWIR1

    @SWIR1.setter
    def SWIR1(self, value):
        self._SWIR1 = self.to_xarray(value)

    @property
    def SWIR2(self):
        return self._SWIR2

    @SWIR2.setter
    def SWIR2(self, value):
        self._SWIR2 = self.to_xarray(value)

    @property
    def TIR1(self):
        return self._TIR1

    @TIR1.setter
    def TIR1(self, value):
        self._TIR1 = self.to_xarray(value)

    @property
    def TIR2(self):
        return self._TIR2

    @TIR2.setter
    def TIR2(self, value):
        self._TIR2 = self.to_xarray(value)

    @property
    def HV(self):
        return self._HV

    @HV.setter
    def HV(self, value):
        self._HV = self.to_xarray(value)

    @property
    def VH(self):
        return self._VH

    @VH.setter
    def VH(self, value):
        self._VH = self.to_xarray(value)

    @property
    def HH(self):
        return self._HH

    @HH.setter
    def HH(self, value):
        self._HH = self.to_xarray(value)

    @property
    def VV(self):
        return self._VV

    @VV.setter
    def VV(self, value):
        self._VV = self.to_xarray(value)

    @property
    def dictarray(self):
        return self._dictarray

    @dictarray.setter
    def dictarray(self, value):
        dictarray = OrderedDict()
        for _, value in self.ref_rev.items():
            obj = eval("self.%s" % value)
            if isinstance(obj, xr.DataArray):
                dictarray[value] = obj

        if self.align:
            resdata = {k: v.rio.resolution()[0] for k, v in dictarray.items()}
            reftensor = dictarray[min(resdata, key=resdata.get)]
            reftensor_res = reftensor.rio.resolution()[0]
            for key, value in dictarray.items():
                if reftensor_res != value.rio.resolution()[0]:
                    dictarray[key] = value.rio.reproject(
                        dst_crs=reftensor.rio.crs,
                        shape=reftensor.shape,
                        transform=reftensor.rio.transform(),
                        resampling=Resampling.nearest,
                    )
                else:
                    dictarray[key] = value
        self._dictarray = dictarray

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, value):
        bounds = set([x.rio.bounds() for x in self.dictarray.values()])
        if len(bounds) > 1:
            raise ValueError("All bands must have the same bounds.")
        self._bounds = bounds.pop()

    def to_dict(self):
        return self.dictarray

    def to_list(self):
        return list(self.dictarray.values())

    def resolution(self):
        dict_res = OrderedDict()
        for key, value in self.dictarray.items():
            dict_res[key] = value.rio.resolution()[0]
        return dict_res

    def rasterbase(self):
        res = self.resolution()
        # Obtain the raster resolution for each band.
        raster_res = [value for key, value in res.items()]

        # Select the raster with the lowest resolution.
        rasterbase_res_index = np.array(raster_res).argmin()
        rasterbase_res_band = list(res.keys())[rasterbase_res_index]
        return self.dictarray[rasterbase_res_band]

    def rasterbase_name(self):
        res = self.resolution()
        # Obtain the raster resolution for each band.
        raster_res = [value for key, value in res.items()]

        # Select the raster with the lowest resolution.
        rasterbase_res_index = np.array(raster_res).argmin()
        return list(res.keys())[rasterbase_res_index]

    def message(self, object: xr.DataArray, band: str, order: int):
        if isinstance(object, xr.DataArray):
            equal_space = 13
            white_space = " " * (equal_space - len("%s [%02d]" % (band, 0)))
            resraster = object.rio.resolution()[0]
            band = color.BLUE + color.BOLD + band + color.END
            order = color.BLUE + color.BOLD + ("%02d" % order) + color.END
            resolution = color.BLUE + color.BOLD + ("%0.1f" % resraster) + color.END

            shp = (
                band,
                white_space,
                order,
                object.shape[0],
                object.shape[1],
                resolution,
            )
            return "%s %s[%s]: <xarray.DataArray (y: %s, x: %s, res: %s)> \n" % shp
        return ""

    def __repr__(self) -> str:
        msg = ""
        for key, value in self.ref_rev.items():
            obj = eval("self.%s" % value)
            if isinstance(obj, xr.DataArray):
                msg += self.message(self.dictarray[value], value, key)
        return msg



import pathlib
import gdown

import numpy as np

from maskay.tensorflow import Module
from maskay.utils import get_models_path


class CDFCNNrgbi(Module):
    def __init__(self):
        super().__init__()
        self.model = model_setup()

    def forward(self, x):
        return self.model(x)

    def inProcessing(self, tensor: np.ndarray):
        # If all the pixels are zero skip the run and outProcessing.
        if np.sum(tensor) == 0:
            shp = tensor.shape
            tensor = np.zeros(
                (shp[0], 1, shp[1], shp[2])
            ) # 1 is the number of the output classes
            return [tensor]
        return tensor / 10000

    def outProcessing(self, tensor: np.ndarray):
        return (tensor * 10000).astype(np.int16)


## Auxiliary functions
def model_setup():
    # Check if packages are installed
    is_external_package_installed = []

    try:
        import tensorflow as tf
    except ImportError:
        is_external_package_installed.append("tensorflow")

    if is_external_package_installed != []:
        nopkgs = ', '.join(is_external_package_installed)
        raise ImportError(
            f"Please install the following packages: {nopkgs}."
        )

    def conv_blocks(
        ip_,
        nfilters,
        axis_batch_norm,
        reg,
        name,
        batch_norm,
        remove_bias_if_batch_norm=False,
        dilation_rate=(1, 1),
    ):
        use_bias = not (remove_bias_if_batch_norm and batch_norm)

        conv = tf.keras.layers.SeparableConv2D(
            nfilters,
            (3, 3),
            padding="same",
            name=name + "_conv_1",
            depthwise_regularizer=reg,
            pointwise_regularizer=reg,
            dilation_rate=dilation_rate,
            use_bias=use_bias,
        )(ip_)

        if batch_norm:
            conv = tf.keras.layers.BatchNormalization(axis=axis_batch_norm, name=name + "_bn_1")(conv)

        conv = tf.keras.layers.Activation("relu", name=name + "_act_1")(conv)

        conv = tf.keras.layers.SeparableConv2D(
            nfilters,
            (3, 3),
            padding="same",
            name=name + "_conv_2",
            use_bias=use_bias,
            dilation_rate=dilation_rate,
            depthwise_regularizer=reg,
            pointwise_regularizer=reg,
        )(conv)

        if batch_norm:
            conv = tf.keras.layers.BatchNormalization(axis=axis_batch_norm, name=name + "_bn_2")(conv)

        return tf.keras.layers.Activation("relu", name=name + "_act_2")(conv)


    def build_unet_model_fun(x_init, weight_decay=0.05, batch_norm=True, final_activation="sigmoid"):

        axis_batch_norm = 3

        reg = tf.keras.regularizers.l2(weight_decay)

        conv1 = conv_blocks(x_init, 32, axis_batch_norm, reg, name="input", batch_norm=batch_norm)

        pool1 = tf.keras.layers.MaxPooling2D(pool_size=(2, 2), name="pooling_1")(conv1)

        conv2 = conv_blocks(pool1, 64, axis_batch_norm, reg, name="pool1", batch_norm=batch_norm)

        pool2 = tf.keras.layers.MaxPooling2D(pool_size=(2, 2), name="pooling_2")(conv2)

        conv3 = conv_blocks(pool2, 128, axis_batch_norm, reg, name="pool2", batch_norm=batch_norm)

        up8 = tf.keras.layers.concatenate(
            [
                tf.keras.layers.Conv2DTranspose(
                    64, (2, 2), strides=(2, 2), padding="same", name="upconv1", kernel_regularizer=reg
                )(conv3),
                conv2,
            ],
            axis=axis_batch_norm,
            name="concatenate_up_1",
        )

        conv8 = conv_blocks(up8, 64, axis_batch_norm, reg, name="up1", batch_norm=batch_norm)

        up9 = tf.keras.layers.concatenate(
            [
                tf.keras.layers.Conv2DTranspose(
                    32, (2, 2), strides=(2, 2), padding="same", name="upconv2", kernel_regularizer=reg
                )(conv8),
                conv1,
            ],
            axis=axis_batch_norm,
            name="concatenate_up_2",
        )

        conv9 = conv_blocks(up9, 32, axis_batch_norm, reg, name="up2", batch_norm=batch_norm)

        conv10 = tf.keras.layers.Conv2D(
            1, (1, 1), kernel_regularizer=reg, name="linear_model", activation=final_activation
        )(conv9)

        return conv10

    def load_model(shape=(None, None), bands_input=4, weight_decay=0.0, final_activation="sigmoid"):
        ip = tf.keras.layers.Input(shape + (bands_input,), name="ip_cloud")
        c11 = tf.keras.layers.Conv2D(bands_input, (1, 1), name="normalization_cloud", trainable=False)
        x_init = c11(ip)
        conv2d10 = build_unet_model_fun(
            x_init, weight_decay=weight_decay, final_activation=final_activation, batch_norm=True
        )
        return tf.keras.models.Model(inputs=[ip], outputs=[conv2d10], name="UNet-Clouds")
        
    filename = (pathlib.Path(get_models_path()) / "CD-FCNN-RGBI.hdf5")
    
    # Download the model if it doesn't exist
    if not filename.is_file():
        # download file using gdown
        url = "https://drive.google.com/uc?id=1TpRE1pPtnBVW7XIJvUOMwCP2nwMmEnij"
        gdown.download(url, filename.as_posix())
    
    # Load the model
    model = load_model(shape=(None, None), bands_input=4)
    model.load_weights(filename.as_posix())
    
    return model




import itertools
from rasterio.enums import Resampling
from collections import OrderedDict
from typing import List
from tqdm import tqdm


import numpy as np
import xarray as xr

from maskay.tensorsat import TensorSat

class MaskayModule:
    def __init__(self, cropsize, overlap, device, batchsize, order, quiet):
        self.cropsize = cropsize
        self.overlap = overlap
        self.device = device
        self.batchsize = batchsize
        self.order = order
        self.quiet = quiet

    @property
    def cropsize(self) -> int:
        return self._cropsize

    @cropsize.setter
    def cropsize(self, value: int):
        if not isinstance(value, int):
            raise TypeError("cropsize must be an integer")
        self._cropsize = value

    @property
    def overlap(self) -> int:
        return self._overlap

    @overlap.setter
    def overlap(self, value: int):
        if not isinstance(value, int):
            raise TypeError("overlap must be an integer")
        self._overlap = value

    @property
    def device(self) -> str:
        return self._device

    @device.setter
    def device(self, value: str):
        if not isinstance(value, str):
            raise TypeError("device must be a string")
        if not value in ["cpu", "cuda"]:
            raise ValueError("device must be either cpu or cuda")
        self._device = value

    @property
    def batchsize(self) -> int:
        return self._batchsize

    @batchsize.setter
    def batchsize(self, value: int):
        if not isinstance(value, int):
            raise TypeError("batchsize must be an integer")
        self._batchsize = value

    @property
    def quiet(self) -> bool:
        return self._quiet

    @quiet.setter
    def quiet(self, value: int):
        if not isinstance(value, int):
            raise TypeError("quiet must be an integer")
        self._quiet = value

    def get_resolution(self, tensor: TensorSat):
        return tensor.resolution()

    def get_cropsize(self, tensor: TensorSat):
        # Base parameters
        cropsize_base = self.cropsize

        # Resolution of the base raster (lowest resolution)
        res_base = tensor.rasterbase().rio.resolution()[0]

        # Resolution of each raster
        res_arrays = self.get_resolution(tensor)

        # Estimate zero-coordinate from each image patch
        dict_cropsize = OrderedDict()
        for key, _ in tensor.dictarray.items():
            factor = res_base / res_arrays[key]
            dict_cropsize[key] = int(np.ceil(cropsize_base * factor))
        return dict_cropsize

    def get_overlap(self, tensor: TensorSat):
        # Base parameters
        overlap_base = self.overlap
        res_base = tensor.rasterbase().rio.resolution()[0]

        # Resolution of each raster
        res_arrays = self.get_resolution(tensor)

        # Estimate zero-coordinate from each image patch
        dict_cropsize = OrderedDict()
        for key, _ in tensor.dictarray.items():
            factor = res_base / res_arrays[key]
            dict_cropsize[key] = int(np.ceil(overlap_base * factor))
        return dict_cropsize

    def get_ips(self, zerocoord: OrderedDict):
        return len(list(zerocoord.values())[0])

    def _MagickCrop(self, tensor: TensorSat):
        # Get the cropsize and overlap for each raster
        tensor_cropsize = self.get_cropsize(tensor)
        tensor_overlap = self.get_overlap(tensor)

        # Estimate zero-coordinate from each image patch
        dict_mrs = OrderedDict()
        for key, value in tensor.dictarray.items():
            cropsize_array = tensor_cropsize[key]
            overlap_array = tensor_overlap[key]
            mrs = self._Crop(
                tensor=value, cropsize=cropsize_array, overlap=overlap_array
            )
            dict_mrs[key] = mrs
        return dict_mrs

    def _Crop(self, tensor: xr.DataArray, cropsize: int, overlap: int):
        # Select the raster with the lowest resolution
        tshp = tensor.shape

        # if the image is too small, return (0, 0)
        if (tshp[0] < cropsize) and (tshp[1] < cropsize):
            return [(0, 0)]

        # Define relative coordinates.
        xmn, xmx, ymn, ymx = (0, tshp[0], 0, tshp[1])

        if overlap > cropsize:
            raise ValueError("The overlap must be smaller than the cropsize")

        xrange = np.arange(xmn, xmx, (cropsize - overlap))
        yrange = np.arange(ymn, ymx, (cropsize - overlap))

        # If there is negative values in the range, change them by zero.
        xrange[xrange < 0] = 0
        yrange[yrange < 0] = 0

        # Remove the last element if it is outside the tensor
        xrange = xrange[xrange - (tshp[0] - cropsize) <= 0]
        yrange = yrange[yrange - (tshp[1] - cropsize) <= 0]

        # If the last element is not (tshp[1] - cropsize) add it!
        if xrange[-1] != (tshp[0] - cropsize):
            xrange = np.append(xrange, tshp[0] - cropsize)
        if yrange[-1] != (tshp[1] - cropsize):
            yrange = np.append(yrange, tshp[1] - cropsize)

        # Create all the relative coordinates
        mrs = list(itertools.product(xrange, yrange))

        return mrs

    def _align(self, reftensor: xr.DataArray, tensor: xr.DataArray):
        # Select a raster base based on the lowest resolution.
        return tensor.rio.reproject(
            dst_crs=reftensor.rio.crs,
            shape=reftensor.shape,
            transform=reftensor.rio.transform(),
            resampling=Resampling.nearest,
        )

    def _MagickGather(self, outensor: np.ndarray, batch_step: List):
        """Gather the image patches and merge them into a single tensor.

        Args:
            tensorprob (np.ndarray): A tensor with shape (B, C, H, W).
            outensor (np.ndarray): A tensor with shape (C, H, W).
            quiet (bool, optional): If True, do not print any messages. Defaults to False.

        Returns:
            np.ndarray: A np.ndarray with shape (C, H, W) with the image patches.
        """

        # Get CropMagick properties
        xmin, xmax = (0, outensor.shape[-1])  # X borders of the output tensor
        ymin, ymax = (0, outensor.shape[-2])  # Y borders of the output tensor

        container = list()
        for coord in batch_step:
            #  Define X dimension pixels
            if coord[0] == xmin:
                Xmin = coord[0]
                XIPmin = 0
            else:
                Xmin = coord[0] + self.overlap // 2
                XIPmin = self.overlap // 2

            if (coord[0] + self.cropsize) == xmax:
                Xmax = coord[0] + self.cropsize
                XIPmax = self.cropsize
            else:
                Xmax = coord[0] + self.cropsize - self.overlap // 2
                XIPmax = self.cropsize - self.overlap // 2

            #  Define Y dimension pixels
            if coord[1] == ymin:
                Ymin = coord[1]
                YIPmin = 0
            else:
                Ymin = coord[1] + self.overlap // 2
                YIPmin = self.overlap // 2

            if (coord[1] + self.cropsize) == ymax:
                Ymax = coord[1] + self.cropsize
                YIPmax = self.cropsize
            else:
                Ymax = coord[1] + self.cropsize - self.overlap // 2
                YIPmax = self.cropsize - self.overlap // 2

            # Put the IP tensor in the output tensor
            container.append(
                {
                    "outensor": [(Xmin, Ymin), (Xmax, Ymax)],
                    "ip": [(XIPmin, YIPmin), (XIPmax, YIPmax)],
                }
            )
        return container
    
    def _predict(self, tensor: TensorSat):
        # Obtain the zero coordinate to create an IP
        zero_coord = self._MagickCrop(tensor)
        
        # Number of image patches (IPs)
        IPslen = self.get_ips(zero_coord)

        # Get the cropsize for each raster
        tensor_cropsize = self.get_cropsize(tensor)

        # Raster ref (lowest resolution)
        rbase = tensor.rasterbase()
        rbase_name = tensor.rasterbase_name()

        # Create outensor
        outensor = None
                
        batch_iter = range(0, IPslen, self.batchsize)
        for index in tqdm(batch_iter, disable=self.quiet):
            batched_IP = list()
            zerocoords = list()
            for index2 in range(index * self.batchsize, (index + 1) * self.batchsize):
                # Container to create a full IP with all bands with the same resolution
                IP = list()

                # Reference raster IP
                bmrx, bmry = zero_coord[rbase_name][index2]
                rbase_ip = rbase[
                    bmrx : (bmrx + self.cropsize), bmry : (bmry + self.cropsize)
                ]
                base_cropsize = tensor_cropsize[rbase_name]

                for key, _ in zero_coord.items():
                    # Select the zero coordinate
                    mrx, mry = zero_coord[key][index2]

                    # Obtain the specific cropsize for each raster
                    cropsize = tensor_cropsize[key]

                    # Crop the full raster using specific coordinates
                    tensorIP = tensor.dictarray[key][
                        mrx : (mrx + cropsize), mry : (mry + cropsize)
                    ]
                    
                    # Resample the raster to the reference raster
                    if base_cropsize != cropsize:
                        tensorIP = self._align(rbase_ip, tensorIP)
                    
                    # Append the IP to the container
                    IP.append(tensorIP.to_numpy())

                # Stack the IP
                IP = np.stack(IP, axis=0)
                
                # Append the IP to the batch
                zerocoords.append((bmrx, bmry))
                batched_IP.append(IP)

            # Stack the batch
            batched_IP = np.stack(batched_IP, axis=0)
            
            # Change order of the batched_IP
            if self.order == "BHWC":
                batched_IP = np.moveaxis(batched_IP, 1, -1)
            
            # Run the preprocessing
            batched_IP = self.inProcessing(batched_IP)
            
            if not isinstance(batched_IP, list):
                # Run the model
                batched_IP = self._run(batched_IP)
                
                # Run the postprocessing
                batched_IP = self.outProcessing(batched_IP)
            else:
                if outensor is not None:
                    batched_IP = batched_IP[0].astype(dtype)
                # If the outensor is not created yet, we do not
                # know the dtype of the output tensor
                else:
                    # Run the model
                    batched_IP = self._run(batched_IP[0])
                    
                    # Run the postprocessing
                    batched_IP = self.outProcessing(batched_IP)
                                                                        
            # If is the first iteration, create the output tensor
            if outensor is None:
                classes = batched_IP.shape[1]
                dtype = batched_IP.dtype
                outensor = np.zeros(
                    shape=(classes, rbase.shape[0], rbase.shape[1]),
                    dtype=dtype
                )

            # Copy the IP values in the outputtensor
            gather_zerocoord = self._MagickGather(outensor, zerocoords)
            
            for index3, zcoords in enumerate(gather_zerocoord):
                # Coordinates to copy the IP
                (Xmin, Ymin), (Xmax, Ymax) = zcoords["outensor"]
                (XIPmin, YIPmin), (XIPmax, YIPmax) = zcoords["ip"]
                
                # Copy the IP
                outensor[:, Xmin:Xmax, Ymin:Ymax] = batched_IP[
                    index3, :, XIPmin:XIPmax, YIPmin:YIPmax
                ]

        # Create the output tensor rio xarray object
        xcoord, ycoord = rbase.coords["y"].values, rbase.coords["x"].values
        coords = [np.arange(0, classes), xcoord, ycoord]
        dims = ["band", "y", "x"]

        return (
            xr.DataArray(outensor, coords=coords, dims=dims)
            .rio.write_nodata(-999)
            .rio.write_transform(rbase.rio.transform())
        )
